# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from PIL import Image
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.blip import BlipProcessor as _BlipProcessor
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import cached_path
from unitorch.cli import WriterOutputs
from unitorch.cli.models import (
    TensorsInputs,
    GenerationOutputs,
    GenerationTargets,
)
from unitorch.cli.models.blip import pretrained_blip_infos


class BlipProcessor(_BlipProcessor):
    """Processor for the BLIP model."""

    def __init__(
        self,
        vocab_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
        position_start_id: Optional[int] = 0,
    ):
        """
        Initialize BlipProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            vision_config_path (str): The path to the vision configuration file.
            max_seq_length (int, optional): The maximum sequence length. Defaults to 128.
            max_gen_seq_length (int, optional): The maximum generated sequence length. Defaults to 48.
            position_start_id (int, optional): The start position ID. Defaults to 0.
        """
        super().__init__(
            vocab_path=vocab_path,
            vision_config_path=vision_config_path,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
            position_start_id=position_start_id,
        )

    @classmethod
    @add_default_section_for_init("core/process/blip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipProcessor from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the processor parameters.
        """
        config.set_default_section("core/process/blip")
        pretrained_name = config.getoption(
            "pretrained_name", "blip-image-captioning-base"
        )

        vocab_path = config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        vision_config_path = config.getoption("vision_config_path", None)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "vision_config"),
        )
        vision_config_path = cached_path(vision_config_path)

        return {
            "vocab_path": vocab_path,
            "vision_config_path": vision_config_path,
        }

    @register_process("core/process/blip/text_classification")
    def _text_classification(
        self,
        text: str,
        max_seq_length: Optional[int] = None,
    ):
        """
        Perform text classification using the BlipProcessor.

        Args:
            text (str): The input text.
            max_seq_length (int, optional): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The tensor inputs.
        """
        outputs = super().text_classification(
            text=text,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            position_ids=outputs.position_ids,
        )

    @register_process("core/process/blip/image_classification")
    def _image_classification(
        self,
        image: Union[Image.Image, str],
    ):
        """
        Perform image classification using the BlipProcessor.

        Args:
            image (Union[Image.Image, str]): The input image.

        Returns:
            TensorsInputs: The tensor inputs.
        """
        if isinstance(image, str):
            image = Image.open(image)
        outputs = super().image_classification(image=image)
        return TensorsInputs(pixel_values=outputs.pixel_values)

    @register_process("core/process/blip/classification")
    def _classification(
        self,
        text: str,
        image: Union[Image.Image, str],
        max_seq_length: Optional[int] = None,
    ):
        """
        Perform text and image classification using the BlipProcessor.

        Args:
            text (str): The input text.
            image (Union[Image.Image, str]): The input image.
            max_seq_length (int, optional): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The tensor inputs.
        """
        if isinstance(image, str):
            image = Image.open(image)

        outputs = super().classification(
            text=text,
            image=image,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            position_ids=outputs.position_ids,
            pixel_values=outputs.pixel_values,
        )

    @register_process("core/process/blip/generation")
    def _generation(
        self,
        text: str,
        image: Union[Image.Image, str],
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Perform text and image generation using the BlipProcessor.

        Args:
            text (str): The input text.
            image (Union[Image.Image, str]): The input image.
            max_gen_seq_length (int, optional): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The tensor inputs.
        """
        if isinstance(image, str):
            image = Image.open(image)

        outputs = super().generation(
            text=text,
            image=image,
            max_gen_seq_length=max_gen_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            pixel_values=outputs.pixel_values,
        ), GenerationTargets(
            refs=outputs.input_ids_label,
            masks=outputs.attention_mask_label,
        )

    @register_process("core/process/blip/generation/labels")
    def _generation_labels(
        self,
        text: str,
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Generate labels using the BlipProcessor.

        Args:
            text (str): The input text.
            max_gen_seq_length (int, optional): The maximum generated sequence length. Defaults to None.

        Returns:
            GenerationTargets: The generation targets.
        """
        outputs = super().generation_labels(
            text=text,
            max_gen_seq_length=max_gen_seq_length,
        )
        return GenerationTargets(
            refs=outputs.input_ids,
            masks=outputs.attention_mask,
        )

    @register_process("core/postprocess/blip/detokenize")
    def _detokenize(
        self,
        outputs: GenerationOutputs,
    ):
        """
        Detokenize generated outputs using the BlipProcessor.

        Args:
            outputs (GenerationOutputs): The generation outputs.

        Returns:
            WriterOutputs: The writer outputs.
        """
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.sequences.shape[0]

        decoded = super().detokenize(sequences=outputs.sequences)
        results["decoded"] = decoded
        return WriterOutputs(results)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import List, Tuple

import cv2
import numpy as np
import onnxruntime as ort


def preprocess(
    img: np.ndarray, out_bbox, input_size: Tuple[int, int] = (192, 256)
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Do preprocessing for RTMPose model inference.

    Args:
        img (np.ndarray): Input image in shape.
        input_size (tuple): Input image size in shape (w, h).

    Returns:
        tuple:
        - resized_img (np.ndarray): Preprocessed image.
        - center (np.ndarray): Center of image.
        - scale (np.ndarray): Scale of image.
    """
    # get shape of image
    img_shape = img.shape[:2]
    out_img, out_center, out_scale = [], [], []
    if len(out_bbox) == 0:
        out_bbox = [[0, 0, img_shape[1], img_shape[0]]]
    for i in range(len(out_bbox)):
        x0 = out_bbox[i][0]
        y0 = out_bbox[i][1]
        x1 = out_bbox[i][2]
        y1 = out_bbox[i][3]
        bbox = np.array([x0, y0, x1, y1])

        # get center and scale
        center, scale = bbox_xyxy2cs(bbox, padding=1.25)

        # do affine transformation
        resized_img, scale = top_down_affine(input_size, scale, center, img)

        # normalize image
        mean = np.array([123.675, 116.28, 103.53])
        std = np.array([58.395, 57.12, 57.375])
        resized_img = (resized_img - mean) / std

        out_img.append(resized_img)
        out_center.append(center)
        out_scale.append(scale)

    return out_img, out_center, out_scale


def inference(sess: ort.InferenceSession, img: np.ndarray) -> np.ndarray:
    """Inference RTMPose model.

    Args:
        sess (ort.InferenceSession): ONNXRuntime session.
        img (np.ndarray): Input image in shape.

    Returns:
        outputs (np.ndarray): Output of RTMPose model.
    """
    all_out = []
    # build input
    for i in range(len(img)):
        input = [img[i].transpose(2, 0, 1)]

        # build output
        sess_input = {sess.get_inputs()[0].name: input}
        sess_output = []
        for out in sess.get_outputs():
            sess_output.append(out.name)

        # run model
        outputs = sess.run(sess_output, sess_input)
        all_out.append(outputs)

    return all_out


def postprocess(
    outputs: List[np.ndarray],
    model_input_size: Tuple[int, int],
    center: Tuple[int, int],
    scale: Tuple[int, int],
    simcc_split_ratio: float = 2.0,
) -> Tuple[np.ndarray, np.ndarray]:
    """Postprocess for RTMPose model output.

    Args:
        outputs (np.ndarray): Output of RTMPose model.
        model_input_size (tuple): RTMPose model Input image size.
        center (tuple): Center of bbox in shape (x, y).
        scale (tuple): Scale of bbox in shape (w, h).
        simcc_split_ratio (float): Split ratio of simcc.

    Returns:
        tuple:
        - keypoints (np.ndarray): Rescaled keypoints.
        - scores (np.ndarray): Model predict scores.
    """
    all_key = []
    all_score = []
    for i in range(len(outputs)):
        # use simcc to decode
        simcc_x, simcc_y = outputs[i]
        keypoints, scores = decode(simcc_x, simcc_y, simcc_split_ratio)

        # rescale keypoints
        keypoints = keypoints / model_input_size * scale[i] + center[i] - scale[i] / 2
        all_key.append(keypoints[0])
        all_score.append(scores[0])

    return np.array(all_key), np.array(all_score)


def bbox_xyxy2cs(
    bbox: np.ndarray, padding: float = 1.0
) -> Tuple[np.ndarray, np.ndarray]:
    """Transform the bbox format from (x,y,w,h) into (center, scale)

    Args:
        bbox (ndarray): Bounding box(es) in shape (4,) or (n, 4), formatted
            as (left, top, right, bottom)
        padding (float): BBox padding factor that will be multilied to scale.
            Default: 1.0

    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32]: Center (x, y) of the bbox in shape (2,) or
            (n, 2)
        - np.ndarray[float32]: Scale (w, h) of the bbox in shape (2,) or
            (n, 2)
    """
    # convert single bbox from (4, ) to (1, 4)
    dim = bbox.ndim
    if dim == 1:
        bbox = bbox[None, :]

    # get bbox center and scale
    x1, y1, x2, y2 = np.hsplit(bbox, [1, 2, 3])
    center = np.hstack([x1 + x2, y1 + y2]) * 0.5
    scale = np.hstack([x2 - x1, y2 - y1]) * padding

    if dim == 1:
        center = center[0]
        scale = scale[0]

    return center, scale


def _fix_aspect_ratio(bbox_scale: np.ndarray, aspect_ratio: float) -> np.ndarray:
    """Extend the scale to match the given aspect ratio.

    Args:
        scale (np.ndarray): The image scale (w, h) in shape (2, )
        aspect_ratio (float): The ratio of ``w/h``

    Returns:
        np.ndarray: The reshaped image scale in (2, )
    """
    w, h = np.hsplit(bbox_scale, [1])
    bbox_scale = np.where(
        w > h * aspect_ratio,
        np.hstack([w, w / aspect_ratio]),
        np.hstack([h * aspect_ratio, h]),
    )
    return bbox_scale


def _rotate_point(pt: np.ndarray, angle_rad: float) -> np.ndarray:
    """Rotate a point by an angle.

    Args:
        pt (np.ndarray): 2D point coordinates (x, y) in shape (2, )
        angle_rad (float): rotation angle in radian

    Returns:
        np.ndarray: Rotated point in shape (2, )
    """
    sn, cs = np.sin(angle_rad), np.cos(angle_rad)
    rot_mat = np.array([[cs, -sn], [sn, cs]])
    return rot_mat @ pt


def _get_3rd_point(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """To calculate the affine matrix, three pairs of points are required. This
    function is used to get the 3rd point, given 2D points a & b.

    The 3rd point is defined by rotating vector `a - b` by 90 degrees
    anticlockwise, using b as the rotation center.

    Args:
        a (np.ndarray): The 1st point (x,y) in shape (2, )
        b (np.ndarray): The 2nd point (x,y) in shape (2, )

    Returns:
        np.ndarray: The 3rd point.
    """
    direction = a - b
    c = b + np.r_[-direction[1], direction[0]]
    return c


def get_warp_matrix(
    center: np.ndarray,
    scale: np.ndarray,
    rot: float,
    output_size: Tuple[int, int],
    shift: Tuple[float, float] = (0.0, 0.0),
    inv: bool = False,
) -> np.ndarray:
    """Calculate the affine transformation matrix that can warp the bbox area
    in the input image to the output size.

    Args:
        center (np.ndarray[2, ]): Center of the bounding box (x, y).
        scale (np.ndarray[2, ]): Scale of the bounding box
            wrt [width, height].
        rot (float): Rotation angle (degree).
        output_size (np.ndarray[2, ] | list(2,)): Size of the
            destination heatmaps.
        shift (0-100%): Shift translation ratio wrt the width/height.
            Default (0., 0.).
        inv (bool): Option to inverse the affine transform direction.
            (inv=False: src->dst or inv=True: dst->src)

    Returns:
        np.ndarray: A 2x3 transformation matrix
    """
    shift = np.array(shift)
    src_w = scale[0]
    dst_w = output_size[0]
    dst_h = output_size[1]

    # compute transformation matrix
    rot_rad = np.deg2rad(rot)
    src_dir = _rotate_point(np.array([0.0, src_w * -0.5]), rot_rad)
    dst_dir = np.array([0.0, dst_w * -0.5])

    # get four corners of the src rectangle in the original image
    src = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale * shift
    src[1, :] = center + src_dir + scale * shift
    src[2, :] = _get_3rd_point(src[0, :], src[1, :])

    # get four corners of the dst rectangle in the input image
    dst = np.zeros((3, 2), dtype=np.float32)
    dst[0, :] = [dst_w * 0.5, dst_h * 0.5]
    dst[1, :] = np.array([dst_w * 0.5, dst_h * 0.5]) + dst_dir
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        warp_mat = cv2.getAffineTransform(np.float32(dst), np.float32(src))
    else:
        warp_mat = cv2.getAffineTransform(np.float32(src), np.float32(dst))

    return warp_mat


def top_down_affine(
    input_size: dict, bbox_scale: dict, bbox_center: dict, img: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the bbox image as the model input by affine transform.

    Args:
        input_size (dict): The input size of the model.
        bbox_scale (dict): The bbox scale of the img.
        bbox_center (dict): The bbox center of the img.
        img (np.ndarray): The original image.

    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32]: img after affine transform.
        - np.ndarray[float32]: bbox scale after affine transform.
    """
    w, h = input_size
    warp_size = (int(w), int(h))

    # reshape bbox to fixed aspect ratio
    bbox_scale = _fix_aspect_ratio(bbox_scale, aspect_ratio=w / h)

    # get the affine matrix
    center = bbox_center
    scale = bbox_scale
    rot = 0
    warp_mat = get_warp_matrix(center, scale, rot, output_size=(w, h))

    # do affine transform
    img = cv2.warpAffine(img, warp_mat, warp_size, flags=cv2.INTER_LINEAR)

    return img, bbox_scale


def get_simcc_maximum(
    simcc_x: np.ndarray, simcc_y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Get maximum response location and value from simcc representations.

    Note:
        instance number: N
        num_keypoints: K
        heatmap height: H
        heatmap width: W

    Args:
        simcc_x (np.ndarray): x-axis SimCC in shape (K, Wx) or (N, K, Wx)
        simcc_y (np.ndarray): y-axis SimCC in shape (K, Wy) or (N, K, Wy)

    Returns:
        tuple:
        - locs (np.ndarray): locations of maximum heatmap responses in shape
            (K, 2) or (N, K, 2)
        - vals (np.ndarray): values of maximum heatmap responses in shape
            (K,) or (N, K)
    """
    N, K, Wx = simcc_x.shape
    simcc_x = simcc_x.reshape(N * K, -1)
    simcc_y = simcc_y.reshape(N * K, -1)

    # get maximum value locations
    x_locs = np.argmax(simcc_x, axis=1)
    y_locs = np.argmax(simcc_y, axis=1)
    locs = np.stack((x_locs, y_locs), axis=-1).astype(np.float32)
    max_val_x = np.amax(simcc_x, axis=1)
    max_val_y = np.amax(simcc_y, axis=1)

    # get maximum value across x and y axis
    mask = max_val_x > max_val_y
    max_val_x[mask] = max_val_y[mask]
    vals = max_val_x
    locs[vals <= 0.0] = -1

    # reshape
    locs = locs.reshape(N, K, 2)
    vals = vals.reshape(N, K)

    return locs, vals


def decode(
    simcc_x: np.ndarray, simcc_y: np.ndarray, simcc_split_ratio
) -> Tuple[np.ndarray, np.ndarray]:
    """Modulate simcc distribution with Gaussian.

    Args:
        simcc_x (np.ndarray[K, Wx]): model predicted simcc in x.
        simcc_y (np.ndarray[K, Wy]): model predicted simcc in y.
        simcc_split_ratio (int): The split ratio of simcc.

    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32]: keypoints in shape (K, 2) or (n, K, 2)
        - np.ndarray[float32]: scores in shape (K,) or (n, K)
    """
    keypoints, scores = get_simcc_maximum(simcc_x, simcc_y)
    keypoints /= simcc_split_ratio

    return keypoints, scores


def inference_pose(session, out_bbox, oriImg):
    h, w = session.get_inputs()[0].shape[2:]
    model_input_size = (w, h)
    resized_img, center, scale = preprocess(oriImg, out_bbox, model_input_size)
    outputs = inference(session, resized_img)
    keypoints, scores = postprocess(outputs, model_input_size, center, scale)

    return keypoints, scores



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_siglip_infos = {
    "siglip-base-patch16-224": {
        "config": hf_endpoint_url(
            "/google/siglip-base-patch16-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/siglip-base-patch16-224/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/google/siglip-base-patch16-224/resolve/main/spiece.model"
        ),
        "weight": hf_endpoint_url(
            "/google/siglip-base-patch16-224/resolve/main/model.safetensors"
        ),
    },
    "siglip-so400m-patch14-384": {
        "config": hf_endpoint_url(
            "/google/siglip-so400m-patch14-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/siglip-so400m-patch14-384/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/google/siglip-so400m-patch14-384/resolve/main/spiece.model"
        ),
        "weight": hf_endpoint_url(
            "/google/siglip-so400m-patch14-384/resolve/main/model.safetensors"
        ),
    },
    "siglip2-base-patch16-224": {
        "config": hf_endpoint_url(
            "/google/siglip2-base-patch16-224/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/siglip2-base-patch16-224/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/google/siglip2-base-patch16-224/resolve/main/tokenizer.model"
        ),
        "weight": hf_endpoint_url(
            "/google/siglip2-base-patch16-224/resolve/main/model.safetensors"
        ),
    },
    "siglip2-so400m-patch14-384": {
        "config": hf_endpoint_url(
            "/google/siglip2-so400m-patch14-384/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/siglip2-so400m-patch14-384/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/google/siglip2-so400m-patch14-384/resolve/main/tokenizer.model"
        ),
        "weight": hf_endpoint_url(
            "/google/siglip2-so400m-patch14-384/resolve/main/model.safetensors"
        ),
    },
    "siglip2-so400m-patch16-512": {
        "config": hf_endpoint_url(
            "/google/siglip2-so400m-patch16-512/resolve/main/config.json"
        ),
        "vision_config": hf_endpoint_url(
            "/google/siglip2-so400m-patch16-512/resolve/main/preprocessor_config.json"
        ),
        "vocab": hf_endpoint_url(
            "/google/siglip2-so400m-patch16-512/resolve/main/tokenizer.model"
        ),
        "weight": hf_endpoint_url(
            "/google/siglip2-so400m-patch16-512/resolve/main/model.safetensors"
        ),
    },
}

pretrained_siglip_extensions_infos = {}

import unitorch.cli.models.siglip.modeling
import unitorch.cli.models.siglip.processing
from unitorch.cli.models.siglip.modeling import (
    SiglipForPretrain,
    SiglipForClassification,
    SiglipForTextClassification,
    SiglipForImageClassification,
)
from unitorch.cli.models.siglip.processing import SiglipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.roberta.modeling import RobertaForClassification, RobertaForMaskLM
from unitorch.models.roberta.processing import RobertaProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.blip.modeling import (
    BlipForPretrain,
    BlipForClassification,
    BlipForTextClassification,
    BlipForImageClassification,
    BlipForImageCaption,
)
from unitorch.models.blip.processing import BlipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.chinese_clip import (
    ChineseClipForPretrain as _ChineseClipForPretrain,
    ChineseClipForClassification as _ChineseClipForClassification,
    ChineseClipForTextClassification as _ChineseClipForTextClassification,
    ChineseClipForImageClassification as _ChineseClipForImageClassification,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import ClassificationOutputs, LossOutputs
from unitorch.cli.models.chinese_clip import pretrained_chinese_clip_infos


@register_model("core/model/pretrain/chinese_clip")
class ChineseClipForPretrain(_ChineseClipForPretrain):
    """CLIP model for pretraining."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
        use_all_gather: Optional[bool] = True,
    ):
        """
        Initialize the ClipForPretrain model.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            freeze_base_model (bool, optional): Whether to freeze the base model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
            use_all_gather (bool, optional): Whether to use all_gather operation. Defaults to True.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
            use_all_gather=use_all_gather,
        )

    @classmethod
    @add_default_section_for_init("core/model/pretrain/chinese_clip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ClipForPretrain from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            ClipForPretrain: An instance of the ClipForPretrain model.
        """
        config.set_default_section("core/model/pretrain/chinese_clip")
        pretrained_name = config.getoption(
            "pretrained_name", "chinese-clip-vit-base-patch16"
        )
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)
        use_all_gather = config.getoption("use_all_gather", True)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
            use_all_gather=use_all_gather,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Perform a forward pass through the model.

        Args:
            input_ids (torch.Tensor): Input token IDs.
            pixel_values (torch.Tensor): Input pixel values.
            attention_mask (torch.Tensor, optional): Attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs. Defaults to None.

        Returns:
            LossOutputs: The loss outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        return LossOutputs(loss=outputs)


@register_model("core/model/classification/chinese_clip")
class ChineseClipForClassification(_ChineseClipForClassification):
    """CLIP model for classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize the ClipForClassification model.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of output classes. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/chinese_clip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ClipForClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            ClipForClassification: An instance of the ClipForClassification model.
        """
        config.set_default_section("core/model/classification/chinese_clip")
        pretrained_name = config.getoption(
            "pretrained_name", "chinese-clip-vit-base-patch16"
        )
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Perform a forward pass through the model.

        Args:
            input_ids (torch.Tensor): Input token IDs.
            pixel_values (torch.Tensor): Input pixel values.
            attention_mask (torch.Tensor, optional): Attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs. Defaults to None.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        return ClassificationOutputs(outputs=outputs)


@register_model("core/model/classification/chinese_clip/text")
class ChineseClipForTextClassification(_ChineseClipForTextClassification):
    """CLIP model for text classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize the ClipForTextClassification model.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of output classes. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/chinese_clip/text")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ClipForTextClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            ClipForTextClassification: An instance of the ClipForTextClassification model.
        """
        config.set_default_section("core/model/classification/chinese_clip/text")
        pretrained_name = config.getoption(
            "pretrained_name", "chinese-clip-vit-base-patch16"
        )
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_Truemodel", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    def forward(
        self,
        input_ids=None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Perform a forward pass through the model.

        Args:
            input_ids (torch.Tensor, optional): Input token IDs. Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs. Defaults to None.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        return ClassificationOutputs(outputs=outputs)


@register_model("core/model/classification/chinese_clip/image")
class ChineseClipForImageClassification(_ChineseClipForImageClassification):
    """CLIP model for image classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize the ClipForImageClassification model.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of output classes. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/chinese_clip/image")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ClipForImageClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            ClipForImageClassification: An instance of the ClipForImageClassification model.
        """
        config.set_default_section("core/model/classification/chinese_clip/image")
        pretrained_name = config.getoption(
            "pretrained_name", "chinese-clip-vit-base-patch16"
        )
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_chinese_clip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast(device_type=("cuda" if torch.cuda.is_available() else "cpu"))
    def forward(
        self,
        pixel_values: torch.Tensor,
    ):
        """
        Perform a forward pass through the model.

        Args:
            pixel_values (torch.Tensor): Input pixel values.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(pixel_values=pixel_values)
        return ClassificationOutputs(outputs=outputs)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_xlm_roberta_infos = {
    "xlm-roberta-base": {
        "config": hf_endpoint_url("/xlm-roberta-base/resolve/main/config.json"),
        "vocab": hf_endpoint_url(
            "/xlm-roberta-base/resolve/main/sentencepiece.bpe.model"
        ),
        "weight": hf_endpoint_url("/xlm-roberta-base/resolve/main/pytorch_model.bin"),
    },
    "xlm-roberta-large": {
        "config": hf_endpoint_url("/xlm-roberta-large/resolve/main/config.json"),
        "vocab": hf_endpoint_url(
            "/xlm-roberta-large/resolve/main/sentencepiece.bpe.model"
        ),
        "weight": hf_endpoint_url("/xlm-roberta-large/resolve/main/pytorch_model.bin"),
    },
    "xlm-roberta-large": {
        "config": hf_endpoint_url("/xlm-roberta-large/resolve/main/config.json"),
        "vocab": hf_endpoint_url(
            "/xlm-roberta-large/resolve/main/sentencepiece.bpe.model"
        ),
        "weight": hf_endpoint_url("/xlm-roberta-large/resolve/main/pytorch_model.bin"),
    },
    "xlm-roberta-large-finetuned-conll03-english": {
        "config": hf_endpoint_url(
            "/xlm-roberta-large-finetuned-conll03-english/resolve/main/config.json"
        ),
        "vocab": hf_endpoint_url(
            "/xlm-roberta-large-finetuned-conll03-english/resolve/main/sentencepiece.bpe.model"
        ),
        "weight": hf_endpoint_url(
            "/xlm-roberta-large-finetuned-conll03-english/resolve/main/pytorch_model.bin"
        ),
    },
    "xlm-roberta-xl": {
        "config": hf_endpoint_url("/facebook/xlm-roberta-xl/resolve/main/config.json"),
        "vocab": hf_endpoint_url(
            "/facebook/xlm-roberta-xl/resolve/main/sentencepiece.bpe.model"
        ),
        "weight": hf_endpoint_url(
            "/facebook/xlm-roberta-xl/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.xlm_roberta.modeling
import unitorch.cli.models.xlm_roberta.processing
from unitorch.cli.models.xlm_roberta.modeling import (
    XLMRobertaForClassification,
    XLMRobertaXLForClassification,
)
from unitorch.cli.models.xlm_roberta.processing import XLMRobertaProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_t5_infos = {
    "t5-base": {
        "config": hf_endpoint_url("/t5-base/resolve/main/config.json"),
        "vocab": hf_endpoint_url("/t5-base/resolve/main/spiece.model"),
        "weight": hf_endpoint_url("/t5-base/resolve/main/pytorch_model.bin"),
    },
    "t5-small": {
        "config": hf_endpoint_url("/t5-small/resolve/main/config.json"),
        "vocab": hf_endpoint_url("/t5-small/resolve/main/spiece.model"),
        "weight": hf_endpoint_url("/t5-small/resolve/main/pytorch_model.bin"),
    },
    "t5-large": {
        "config": hf_endpoint_url("/t5-large/resolve/main/config.json"),
        "vocab": hf_endpoint_url("/t5-large/resolve/main/spiece.model"),
        "weight": hf_endpoint_url("/t5-large/resolve/main/pytorch_model.bin"),
    },
}

import unitorch.cli.models.t5.modeling
import unitorch.cli.models.t5.processing
from unitorch.cli.models.t5.modeling import T5ForGeneration
from unitorch.cli.models.t5.processing import T5Processor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from transformers import BertConfig, BertModel
from unitorch.models import GenericModel


class BertForClassification(GenericModel):
    replace_keys_in_state_dict = {"gamma": "weight", "beta": "bias"}

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the BertForClassification model.

        Args:
            config_path (str): The path to the configuration file.
            num_classes (Optional[int], optional): The number of classes for classification. Defaults to 1.
            gradient_checkpointing (Optional[bool], optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BertConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.bert = BertModel(self.config)
        self.dropout = nn.Dropout(self.config.hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the BertForClassification model.

        Args:
            input_ids (torch.Tensor): The input tensor of token indices.
            attention_mask (torch.Tensor optional): The attention mask tensor. Defaults to None.
            token_type_ids (torch.Tensor optional): The token type IDs tensor. Defaults to None.
            position_ids (torch.Tensor optional): The position IDs tensor. Defaults to None.

        Returns:
            (torch.Tensor):The logits of the model output.
        """
        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.utils import (
    is_diffusers_available,
    is_opencv_available,
)

if is_diffusers_available():
    from diffusers.schedulers import (
        DDIMScheduler,
        DDPMScheduler,
        PNDMScheduler,
        DEISMultistepScheduler,
        DPMSolverSDEScheduler,
        UniPCMultistepScheduler,
        EulerAncestralDiscreteScheduler,
        EulerDiscreteScheduler,
        FlowMatchEulerDiscreteScheduler,
    )

    Schedulers = {
        "DPM++SDE": DPMSolverSDEScheduler,
        "UniPC++": UniPCMultistepScheduler,
        "DEIS++": DEISMultistepScheduler,
        "DDIM": DDIMScheduler,
        "DDPM": DDPMScheduler,
        "PNDM": PNDMScheduler,
        "FlowMatchEuler": FlowMatchEulerDiscreteScheduler,
        "EulerAncestral": EulerAncestralDiscreteScheduler,
        "Euler": EulerDiscreteScheduler,
    }
    import unitorch.cli.pipelines.stable
    import unitorch.cli.pipelines.stable_xl

    import unitorch.cli.pipelines.stable_3
    import unitorch.cli.pipelines.stable_flux

    if is_opencv_available():
        import unitorch.cli.pipelines.wan
else:
    Schedulers = {}

import unitorch.cli.pipelines.blip
import unitorch.cli.pipelines.bloom
import unitorch.cli.pipelines.bria
import unitorch.cli.pipelines.detr
import unitorch.cli.pipelines.dpt
import unitorch.cli.pipelines.grounding_dino
import unitorch.cli.pipelines.segformer
import unitorch.cli.pipelines.llama
import unitorch.cli.pipelines.llava
import unitorch.cli.pipelines.mask2former
import unitorch.cli.pipelines.mistral
import unitorch.cli.pipelines.sam



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from peft import LoraConfig, PeftModelForCausalLM
from transformers.models.clip.modeling_clip import (
    CLIPConfig,
    CLIPTextTransformer,
    CLIPVisionTransformer,
)
from unitorch.models import GenericModel, GenericOutputs
from unitorch.models.peft import (
    PeftModelForSequenceClassification,
    GenericPeftModel,
    PeftWeightLoaderMixin,
)


class ClipForMatching(GenericModel):
    """
    Clip model for pretraining.
    """

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
    ):
        """
        Initializes the ClipForPretrain model.

        Args:
            config_path (str): Path to the model configuration file.
            projection_dim (int, optional): Dimension of the projected embeddings. Defaults to 512.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
            use_all_gather (bool, optional): Whether to use all-gather operation. Defaults to True.
        """
        super().__init__()

        config = CLIPConfig.from_json_file(config_path)
        text_config = config.text_config
        vision_config = config.vision_config

        self.projection_dim = projection_dim

        self.text_embed_dim = text_config.hidden_size
        self.vision_embed_dim = vision_config.hidden_size

        self.text_model = CLIPTextTransformer(text_config)
        self.vision_model = CLIPVisionTransformer(vision_config)

        self.visual_projection = nn.Linear(
            self.vision_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.text_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )

        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor,
        position_ids: torch.Tensor,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        """
        Forward pass of the Clip model.

        Args:
            input_ids (torch.Tensor, optional): Input text token IDs. Defaults to None.
            pixel_values (torch.Tensor, optional): Input image pixel values. Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask for the input. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs for the input tokens. Defaults to None.
            output_attentions (bool, optional): Whether to output attentions. Defaults to None.
            output_hidden_states (bool, optional): Whether to output hidden states. Defaults to None.

        Returns:
            (torch.Tensor):Logits per text.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )

        text_outputs = self.text_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )

        image_embeds = vision_outputs[1]
        image_embeds = self.visual_projection(image_embeds)

        text_embeds = text_outputs[1]
        text_embeds = self.text_projection(text_embeds)

        return (text_embeds, image_embeds)


class ClipForTextMatching(GenericModel):
    """
    Clip model for pretraining.
    """

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
    ):
        """
        Initializes the ClipForPretrain model.

        Args:
            config_path (str): Path to the model configuration file.
            projection_dim (int, optional): Dimension of the projected embeddings. Defaults to 512.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
            use_all_gather (bool, optional): Whether to use all-gather operation. Defaults to True.
        """
        super().__init__()

        config = CLIPConfig.from_json_file(config_path)
        text_config = config.text_config

        self.projection_dim = projection_dim

        self.text_embed_dim = text_config.hidden_size

        self.query_model = CLIPTextTransformer(text_config)
        self.doc_model = CLIPTextTransformer(text_config)

        self.query_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.doc_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )

        self.init_weights()

    def forward(
        self,
        query_input_ids: torch.Tensor,
        query_attention_mask: torch.Tensor,
        query_position_ids: torch.Tensor,
        doc_input_ids: torch.Tensor,
        doc_attention_mask: torch.Tensor,
        doc_position_ids: torch.Tensor,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        """
        Forward pass of the Clip model.

        Args:
            input_ids (torch.Tensor, optional): Input text token IDs. Defaults to None.
            pixel_values (torch.Tensor, optional): Input image pixel values. Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask for the input. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs for the input tokens. Defaults to None.
            output_attentions (bool, optional): Whether to output attentions. Defaults to None.
            output_hidden_states (bool, optional): Whether to output hidden states. Defaults to None.

        Returns:
            (torch.Tensor):Logits per text.
        """

        query_outputs = self.query_model(
            input_ids=query_input_ids,
            attention_mask=query_attention_mask,
            position_ids=query_position_ids,
        )

        query_embeds = query_outputs[1]
        query_embeds = self.query_projection(query_embeds)

        doc_outputs = self.doc_model(
            input_ids=doc_input_ids,
            attention_mask=doc_attention_mask,
            position_ids=doc_position_ids,
        )

        doc_embeds = doc_outputs[1]
        doc_embeds = self.doc_projection(doc_embeds)

        return (query_embeds, doc_embeds)


class ClipLoraForMatching(GenericPeftModel, PeftWeightLoaderMixin):
    prefix_keys_in_state_dict = {
        "^(?!peft_model\.base_model\.model\.).*": "peft_model.base_model.model."
    }
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }
    modules_to_save_checkpoints = ["lora", "classifier"]
    replace_keys_in_peft_state_dict = {
        ".weight": ".base_layer.weight",
        ".bias": ".base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
    ):
        super().__init__()
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.peft_model = PeftModelForSequenceClassification(
            ClipForMatching(config_path, projection_dim=projection_dim),
            self.peft_config,
        )
        self.classifier = nn.Linear(1, 1)

        self.init_weights()
        self.classifier.weight.data.fill_(5.0)

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor,
        position_ids: torch.Tensor,
    ):
        text_embeds, image_embeds = self.peft_model(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )

        text_embeds = text_embeds / text_embeds.norm(dim=-1, keepdim=True)
        image_embeds = image_embeds / image_embeds.norm(dim=-1, keepdim=True)
        scores = torch.sum(text_embeds * image_embeds, dim=-1, keepdim=True)

        outputs = self.classifier(scores)
        return outputs


class ClipLoraForTextMatching(GenericPeftModel, PeftWeightLoaderMixin):
    prefix_keys_in_state_dict = {
        "^(?!peft_model\.base_model\.model\.).*": "peft_model.base_model.model."
    }
    replace_keys_in_state_dict = {
        "q_proj.weight": "q_proj.base_layer.weight",
        "q_proj.bias": "q_proj.base_layer.bias",
        "v_proj.weight": "v_proj.base_layer.weight",
        "v_proj.bias": "v_proj.base_layer.bias",
    }
    modules_to_save_checkpoints = ["lora", "classifier"]
    replace_keys_in_peft_state_dict = {
        ".weight": ".base_layer.weight",
        ".bias": ".base_layer.bias",
    }

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
    ):
        super().__init__()
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        self.peft_model = PeftModelForSequenceClassification(
            ClipForTextMatching(config_path, projection_dim=projection_dim),
            self.peft_config,
        )
        self.classifier = nn.Linear(1, 1)

        self.init_weights()
        self.classifier.weight.data.fill_(5.0)

    def forward(
        self,
        query_input_ids: torch.Tensor,
        query_attention_mask: torch.Tensor,
        query_position_ids: torch.Tensor,
        doc_input_ids: torch.Tensor,
        doc_attention_mask: torch.Tensor,
        doc_position_ids: torch.Tensor,
    ):
        query_embeds, doc_embeds = self.peft_model(
            query_input_ids=query_input_ids,
            query_attention_mask=query_attention_mask,
            query_position_ids=query_position_ids,
            doc_input_ids=doc_input_ids,
            doc_attention_mask=doc_attention_mask,
            doc_position_ids=doc_position_ids,
        )

        query_embeds = query_embeds / query_embeds.norm(dim=-1, keepdim=True)
        doc_embeds = doc_embeds / doc_embeds.norm(dim=-1, keepdim=True)
        scores = torch.sum(query_embeds * doc_embeds, dim=-1, keepdim=True)

        outputs = self.classifier(scores)
        return outputs



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import nested_dict_value
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.pipelines.bloom import BloomForGenerationPipeline
from unitorch.cli.models.bloom import (
    pretrained_bloom_infos,
    pretrained_bloom_extensions_infos,
)
from unitorch.cli.webuis import (
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_lora_layout,
)
from unitorch.cli.webuis import SimpleWebUI


@register_webui("core/webui/bloom")
class BloomWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_bloom_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names,
        ["^bloom-", "^bloomz-"],
    )
    pretrained_extension_names = list(pretrained_bloom_extensions_infos.keys())
    supported_lora_names = matched_pretrained_names(
        pretrained_extension_names, ["^bloom-lora-", "^bloomz-lora-"]
    )

    def __init__(self, config: CoreConfigureParser):
        self._pipe = None
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )
        self.num_loras = 5
        lora_layout_group = create_lora_layout(
            self.supported_lora_names, num_loras=self.num_loras
        )
        loras = lora_layout_group.loras
        lora_layout = lora_layout_group.layout
        lora_params = []
        for lora in loras:
            lora_params += [
                lora.checkpoint,
                lora.weight,
                lora.alpha,
                lora.url,
                lora.file,
            ]

        prompt = create_element("text", "Input Prompt", lines=3)
        generate = create_element("button", "Generate")
        result = create_element("text", "Output Result", lines=3)

        # create blocks
        left = create_column(prompt, lora_layout, generate)
        right = create_column(result)
        iface = create_blocks(pretrain_layout, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(self.start, inputs=[name], outputs=[status], trigger_mode="once")
        stop.click(self.stop, outputs=[status], trigger_mode="once")

        for lora in loras:
            lora.checkpoint.change(
                fn=lambda x: nested_dict_value(
                    pretrained_bloom_extensions_infos, x, "text"
                ),
                inputs=[lora.checkpoint],
                outputs=[lora.text],
            )

        generate.click(
            self.serve,
            inputs=[
                prompt,
                *lora_params,
            ],
            outputs=[result],
            trigger_mode="once",
        )

        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="Bloom", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._name == pretrained_name and self._status == "Running":
            return self._status
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = BloomForGenerationPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def serve(
        self,
        text: str,
        *params,
    ):
        assert self._pipe is not None
        lora_params = params
        lora_checkpoints = lora_params[0::5]
        lora_weights = lora_params[1::5]
        lora_alphas = lora_params[2::5]
        lora_urls = lora_params[3::5]
        lora_files = lora_params[4::5]
        result = self._pipe(
            text,
            lora_checkpoints=lora_checkpoints,
            lora_weights=lora_weights,
            lora_alphas=lora_alphas,
            lora_urls=lora_urls,
            lora_files=lora_files,
        )
        return result



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.dpt import (
    DPTForDepthEstimation as _DPTForDepthEstimation,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import SegmentationOutputs, LossOutputs
from unitorch.cli.models import segmentation_model_decorator
from unitorch.cli.models.dpt import pretrained_dpt_infos


@register_model("core/model/dpt", segmentation_model_decorator)
class DPTForDepthEstimation(_DPTForDepthEstimation):
    def __init__(
        self,
        config_path: str,
    ):
        super().__init__(
            config_path=config_path,
        )

    @classmethod
    @add_default_section_for_init("core/model/dpt")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/dpt")
        pretrained_name = config.getoption("pretrained_name", "dpt-large")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_dpt_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        inst = cls(
            config_path=config_path,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_dpt_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    def forward(
        self,
    ):
        raise NotImplementedError

    @add_default_section_for_function("core/model/dpt")
    @torch.no_grad()
    def segment(
        self,
        pixel_values: torch.Tensor,
    ):
        outputs = super().forward(
            pixel_values=pixel_values,
        )
        max_values = torch.amax(outputs.reshape(len(outputs), -1), dim=1)
        return SegmentationOutputs(
            masks=list([out / value for out, value in zip(outputs, max_values)]),
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import logging
import hashlib
import ast
import abc
import torch
import torch.nn as nn
import configparser
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from copy import copy, deepcopy
from transformers.utils import is_remote_url


# core config class
class CoreConfigureParser(configparser.ConfigParser):
    def __init__(
        self,
        fpath: Optional[str] = "./config.ini",
        params: Optional[Tuple[str]] = [],
    ):
        super().__init__(interpolation=configparser.ExtendedInterpolation())
        self.fpath = fpath
        self.read(fpath)
        for param in params:
            assert len(param) == 3
            k0, k1, v = param
            if not self.has_section(k0):
                self.add_section(k0)
            self.set(k0, k1, str(v))

        self._freeze_section = None

        self._default_section = self.getdefault("core/config", "default_section", None)

    def _getdefault(self, section, option, value=None):
        if not self.has_section(section):
            return value
        if self.has_option(section, option):
            return self.get(section, option)
        return value

    def _ast_replacement(self, node):
        value = node.id
        if value in ("True", "False", "None"):
            return node
        return ast.Str(value)

    def _ast_literal_eval(self, value):
        root = ast.parse(value, mode="eval")
        if isinstance(root.body, ast.BinOp):
            raise ValueError(value)

        for node in ast.walk(root):
            for field, child in ast.iter_fields(node):
                if isinstance(child, list):
                    for index, subchild in enumerate(child):
                        if isinstance(subchild, ast.Name):
                            child[index] = self._ast_replacement(subchild)
                elif isinstance(child, ast.Name):
                    replacement = self._ast_replacement(child)
                    node.__setattr__(field, replacement)
        return ast.literal_eval(root)

    def get(
        self,
        section,
        option,
        raw=False,
        vars=None,
        fallback=configparser._UNSET,
    ):
        value = super().get(
            section,
            option,
            raw=raw,
            vars=vars,
            fallback=fallback,
        )
        if raw:
            return value
        try:
            return self._ast_literal_eval(value)
        except (SyntaxError, ValueError):
            return value

    def set_default_section(self, section):
        self._default_section = section

    def getdefault(self, section, option, value=None):
        return self._getdefault(section, option, value)

    def getoption(self, option, value=None):
        return self.getdefault(self._default_section, option, value)

    def logging(self):
        logging.info("#" * 30, "Config Info".center(20, " "), "#" * 30)
        for sec, item in self.items():
            for k, v in item.items():
                logging.info(
                    sec.rjust(10, " "),
                    ":".center(5, " "),
                    k.ljust(30, " "),
                    ":".center(5, " "),
                    str(v).ljust(30, " "),
                )
        logging.info("#" * 30, "Config Info End".center(20, " "), "#" * 30)

    def save(self, save_path="./config.ini"):
        self.write(open(save_path, "w"))
        return save_path

    def hexsha(self, length=None):
        string = sorted(
            [f"{k}_{kk}_{vv}" for k, v in self.items() for kk, vv in v.items()]
        )
        string = "|".join(string)
        hexsha = hashlib.sha1(string.encode()).hexdigest()
        if length is not None:
            hexsha = hexsha[:length]
        return hexsha

    def __copy__(self):
        setting = [(sec, k, v) for sec in self.sections() for k, v in self[sec].items()]
        return type(self)(self.fpath, setting)

    def __deepcopy__(self, memo):
        setting = [(sec, k, v) for sec in self.sections() for k, v in self[sec].items()]
        return type(self)(self.fpath, setting)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from transformers import CLIPTextConfig, CLIPTextModel, CLIPTextModelWithProjection
from diffusers.schedulers import SchedulerMixin
from diffusers.models import (
    T2IAdapter,
    MultiAdapter,
    UNet2DModel,
    UNet2DConditionModel,
    AutoencoderKL,
)
from diffusers.pipelines import StableDiffusionXLAdapterPipeline
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.diffusers import GenericStableXLModel


class StableXLAdapterForText2ImageGeneration(GenericStableXLModel):
    """
    StableXLAdapter model for text-to-image generation.

    Args:
        config_path (str): Path to the model configuration file.
        text_config_path (str): Path to the text model configuration file.
        text2_config_path (str): Path to the second text model configuration file.
        vae_config_path (str): Path to the VAE model configuration file.
        adapter_configs_path (str): Path to the adapter model configuration file.
        scheduler_config_path (str): Path to the scheduler configuration file.
        quant_config_path (Optional[str]): Path to the quantization configuration file (default: None).
        image_size (Optional[int]): Size of the input image (default: None).
        in_channels (Optional[int]): Number of input channels (default: None).
        out_channels (Optional[int]): Number of output channels (default: None).
        num_train_timesteps (Optional[int]): Number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): Number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        freeze_unet_encoder (Optional[bool]): Whether to freeze the UNet encoder (default: True).
        seed (Optional[int]): Random seed (default: 1123).
    """

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        adapter_configs_path: Union[str, List[str]],
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = False,
        snr_gamma: Optional[float] = 5.0,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            adapter_configs_path=adapter_configs_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_unet_encoder=freeze_unet_encoder,
            snr_gamma=snr_gamma,
            seed=seed,
        )
        self.pipeline = StableDiffusionXLAdapterPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            unet=self.unet,
            adapter=self.adapter,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        add_time_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        adapter_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the model.

        Args:
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            add_time_ids (torch.Tensor): Additional time IDs.
            pixel_values (torch.Tensor): Pixel values.
            adapter_pixel_values (torch.Tensor): Condition pixel values.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).

        Returns:
            torch.Tensor: Loss value.
        """
        prompt_outputs = self.text(
            input_ids,
            # attention_mask,
            output_hidden_states=True,
        )
        prompt_embeds = prompt_outputs.hidden_states[-2]
        prompt2_outputs = self.text2(
            input2_ids,
            # attention2_mask,
            output_hidden_states=True,
        )
        prompt2_embeds = prompt2_outputs.hidden_states[-2]
        prompt_embeds = torch.concat([prompt_embeds, prompt2_embeds], dim=-1)
        pooled_prompt_embeds = prompt2_outputs[0]

        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor

        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.size(0)

        timesteps = torch.randint(
            0,
            self.scheduler.config.num_train_timesteps,
            (batch,),
            device=pixel_values.device,
        ).long()

        noise_latents = self.scheduler.add_noise(
            latents,
            noise,
            timesteps,
        )

        down_block_additional_residuals = self.adapter(adapter_pixel_values)
        down_block_additional_residuals = [
            sample.to(dtype=noise_latents.dtype)
            for sample in down_block_additional_residuals
        ]
        outputs = self.unet(
            noise_latents,
            timesteps,
            prompt_embeds,
            added_cond_kwargs={
                "time_ids": add_time_ids,
                "text_embeds": pooled_prompt_embeds,
            },
            down_block_additional_residuals=down_block_additional_residuals,
        ).sample

        if self.scheduler.config.prediction_type == "v_prediction":
            noise = self.scheduler.get_velocity(latents, noise, timesteps)

        loss = F.mse_loss(outputs, noise, reduction="mean")
        return loss

    def generate(
        self,
        adapter_pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 5.0,
        adapter_conditioning_scale: Optional[Union[float, List[float]]] = None,
    ):
        """
        Generate images using the model.

        Args:
            adapter_pixel_values (torch.Tensor): Condition pixel values.
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            negative_input_ids (torch.Tensor): Negative input IDs.
            negative_input2_ids (torch.Tensor): Negative second input IDs.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): Negative attention mask (default: None).
            negative_attention2_mask (Optional[torch.Tensor]): Negative second attention mask (default: None).
            height (Optional[int]): Height of the generated images (default: 1024).
            width (Optional[int]): Width of the generated images (default: 1024).
            guidance_scale (Optional[float]): Scale for guidance (default: 5.0).

        Returns:
            GenericOutputs: Generated images.
        """
        outputs = self.get_prompt_outputs(
            input_ids=input_ids,
            input2_ids=input2_ids,
            negative_input_ids=negative_input_ids,
            negative_input2_ids=negative_input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
            negative_attention_mask=negative_attention_mask,
            negative_attention2_mask=negative_attention2_mask,
        )
        if adapter_conditioning_scale is None:
            if self.num_adapters == 1:
                adapter_conditioning_scale = 1.0
            else:
                adapter_conditioning_scale = [1.0] * self.num_adapters
        elif not isinstance(adapter_conditioning_scale, list) and self.num_adapters > 1:
            adapter_conditioning_scale = [
                adapter_conditioning_scale
            ] * self.num_adapters

        images = self.pipeline(
            image=adapter_pixel_values
            if self.num_adapters == 1
            else list(adapter_pixel_values.transpose(0, 1)),
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            negative_pooled_prompt_embeds=outputs.negative_pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            num_inference_steps=self.num_infer_timesteps,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            adapter_conditioning_scale=adapter_conditioning_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models.mask2former import (
    Mask2FormerForSegmentation as _Mask2FormerForSegmentation,
)
from unitorch.models.mask2former import Mask2FormerProcessor
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models.mask2former import pretrained_mask2former_infos


class Mask2FormerForSegmentationPipeline(_Mask2FormerForSegmentation):
    def __init__(
        self,
        config_path: str,
        vision_config_path: str,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        enable_cpu_offload: Optional[bool] = True,
        device: Optional[Union[str, int]] = "cpu",
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = Mask2FormerProcessor(
            vision_config_path=vision_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self._enable_cpu_offload = enable_cpu_offload
        if not self._enable_cpu_offload and self._device != "cpu":
            self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/mask2former")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        config.set_default_section("core/pipeline/mask2former")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "mask2former-swin-tiny-ade-semantic"
        )

        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_mask2former_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vision_config_path = vision_config_path or config.getoption(
            "vision_config_path", None
        )
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(
                pretrained_mask2former_infos, pretrained_name, "vision_config"
            ),
        )
        vision_config_path = cached_path(vision_config_path)

        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        device = config.getoption("device", "cpu") if device is None else device
        pretrained_weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_mask2former_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vision_config_path,
            weight_path=weight_path,
            enable_cpu_offload=enable_cpu_offload,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/mask2former")
    def __call__(
        self,
        image: Union[Image.Image, str],
    ):
        if self._enable_cpu_offload:
            self.to(self._device)
        if isinstance(image, str):
            image = Image.open(image).convert("RGB")
        width, height = image.size
        inputs = self.processor.classification(image)
        outputs = self.segment(inputs.pixel_values.unsqueeze(0).to(self._device))
        masks, classes = outputs.masks.sigmoid(), outputs.classes.argmax(dim=-1)
        # resize masks to the original image size with nn.functional.interpolate
        masks = torch.nn.functional.interpolate(
            masks,
            size=(height, width),
            mode="bilinear",
            align_corners=False,
        )
        masks = masks.cpu().numpy()[0]
        classes = classes.cpu().numpy()[0]
        id2label = self.model.config.id2label
        labels = [id2label.get(int(cls), None) for cls in classes]
        if self._enable_cpu_offload:
            self.to("cpu")
            torch.cuda.empty_cache()
        return [
            (mask, label) for mask, label in zip(masks, labels) if label is not None
        ]



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_dinov2_infos = {
    "dinov2-base": {
        "config": hf_endpoint_url("/facebook/dinov2-base/resolve/main/config.json"),
        "vision_config": hf_endpoint_url(
            "/facebook/dinov2-base/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/dinov2-base/resolve/main/pytorch_model.bin"
        ),
    },
    "dinov2-small": {
        "config": hf_endpoint_url("/facebook/dinov2-small/resolve/main/config.json"),
        "vision_config": hf_endpoint_url(
            "/facebook/dinov2-small/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/dinov2-small/resolve/main/pytorch_model.bin"
        ),
    },
    "dinov2-large": {
        "config": hf_endpoint_url("/facebook/dinov2-large/resolve/main/config.json"),
        "vision_config": hf_endpoint_url(
            "/facebook/dinov2-large/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/dinov2-large/resolve/main/pytorch_model.bin"
        ),
    },
    "dinov2-giant": {
        "config": hf_endpoint_url("/facebook/dinov2-giant/resolve/main/config.json"),
        "vision_config": hf_endpoint_url(
            "/facebook/dinov2-giant/resolve/main/preprocessor_config.json"
        ),
        "weight": hf_endpoint_url(
            "/facebook/dinov2-giant/resolve/main/pytorch_model.bin"
        ),
    },
}

import unitorch.cli.models.dinov2.modeling
import unitorch.cli.models.dinov2.processing
from unitorch.cli.models.dinov2.modeling import DinoV2ForImageClassification
from unitorch.cli.models.dinov2.processing import DinoV2Processor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from torch.optim import Adam, SGD
from torch.optim.adamw import AdamW
from transformers import Adafactor
from unitorch.optims.lion import Lion

from unitorch.utils import is_bitsandbytes_available

if is_bitsandbytes_available():
    from bitsandbytes.optim import Adam8bit, AdamW8bit



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models.grounding_dino import (
    GroundingDinoForDetection as _GroundingDinoForDetection,
)
from unitorch.models.grounding_dino import GroundingDinoProcessor
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericScript
from unitorch.cli import register_script
from unitorch.cli.models.grounding_dino import pretrained_grounding_dino_infos


class GroundingDinoForDetectionPipeline(_GroundingDinoForDetection):
    def __init__(
        self,
        config_path: str,
        vocab_path: str,
        vision_config_path: str,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        enable_cpu_offload: Optional[bool] = True,
        device: Optional[Union[str, int]] = "cpu",
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = GroundingDinoProcessor(
            vocab_path=vocab_path,
            vision_config_path=vision_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self._enable_cpu_offload = enable_cpu_offload
        if not self._enable_cpu_offload and self._device != "cpu":
            self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/grounding_dino")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        vision_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        config.set_default_section("core/pipeline/grounding_dino")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "grounding-dino-tiny"
        )

        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(
                pretrained_grounding_dino_infos, pretrained_name, "config"
            ),
        )
        config_path = cached_path(config_path)

        vocab_path = vocab_path or config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(
                pretrained_grounding_dino_infos, pretrained_name, "vocab"
            ),
        )
        vocab_path = cached_path(vocab_path)

        vision_config_path = vision_config_path or config.getoption(
            "vision_config_path", None
        )
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(
                pretrained_grounding_dino_infos, pretrained_name, "vision_config"
            ),
        )
        vision_config_path = cached_path(vision_config_path)

        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        device = config.getoption("device", "cpu") if device is None else device
        pretrained_weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(
                pretrained_grounding_dino_infos, pretrained_name, "weight"
            ),
            check_none=False,
        )

        inst = cls(
            config_path,
            vocab_path,
            vision_config_path,
            weight_path=weight_path,
            enable_cpu_offload=enable_cpu_offload,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/grounding_dino")
    def __call__(
        self,
        text: str,
        image: Union[Image.Image, str],
        text_threshold: Optional[float] = 0.25,
        box_threshold: Optional[float] = 0.25,
    ):
        if self._enable_cpu_offload:
            self.to(self._device)
        inputs = self.processor.detection_inputs(text, image)
        outputs = self.detect(
            inputs.pixel_values.unsqueeze(0).to(self._device),
            inputs.input_ids.unsqueeze(0).to(self._device),
            inputs.token_type_ids.unsqueeze(0).to(self._device),
            inputs.attention_mask.unsqueeze(0).to(self._device),
            norm_bboxes=True,
            text_threshold=text_threshold,
            box_threshold=box_threshold,
        )
        result_image = image.copy()
        bboxes = outputs["bboxes"][0].cpu().numpy()
        scores = outputs["scores"][0].cpu().numpy()
        classes = outputs["classes"][0].cpu().numpy()
        # draw boxes on the result image with class and score
        for bbox, score, tokens in zip(bboxes, scores, classes):
            if score < box_threshold:
                continue
            classid = self.processor.tokenizer.decode(
                [t for t in tokens if t != 0],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
            bbox = bbox * np.array(
                [
                    result_image.width,
                    result_image.height,
                    result_image.width,
                    result_image.height,
                ]
            )
            bbox = list(map(int, bbox))
            result_image = result_image.copy()
            draw = ImageDraw.Draw(result_image)
            draw.rectangle(bbox, outline="red", width=3)
            draw.text((bbox[0] + 5, bbox[1] + 5), f"{classid}", fill="blue")
            draw.text((bbox[0] + 5, bbox[1] + 15), f"{score:.2f}", fill="green")
        if self._enable_cpu_offload:
            self.to("cpu")
            torch.cuda.empty_cache()
        return result_image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import re
import gc
import json
import logging
import torch
import hashlib
import asyncio
import pandas as pd
from PIL import Image
from torch import autocast
from fastapi import APIRouter, UploadFile, File
from fastapi.responses import StreamingResponse
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil
from diffusers.models import ControlNetModel
from diffusers.pipelines import (
    WanImageToVideoPipeline,
)
from unitorch import is_xformers_available
from unitorch.utils import is_remote_url, tensor2vid
from unitorch.models.diffusers import WanForImage2VideoGeneration
from unitorch.models.diffusers import WanProcessor
from unitorch.utils import (
    pop_value,
    nested_dict_value,
    is_bfloat16_available,
    is_cuda_available,
)
from unitorch.cli import (
    cached_path,
    register_fastapi,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli import CoreConfigureParser, GenericFastAPI
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
    load_weight,
)
from unitorch.cli.pipelines import Schedulers
from unitorch.cli.models.diffusion_utils import export_to_video


class WanForImage2VideoFastAPIPipeline(WanForImage2VideoGeneration):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        image_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        vocab_path: str,
        image_process_config_path: str,
        quant_config_path: Optional[str] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        lora_checkpoints: Optional[Union[str, List[str]]] = None,
        lora_weights: Optional[Union[float, List[float]]] = 1.0,
        lora_alphas: Optional[Union[float, List[float]]] = 32,
        device: Optional[Union[str, int]] = "cpu",
        enable_cpu_offload: Optional[bool] = False,
        enable_xformers: Optional[bool] = False,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            image_config_path=image_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
        )
        self.processor = WanProcessor(
            vocab_path=vocab_path,
            image_config_path=image_process_config_path,
            vae_config_path=vae_config_path,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.eval()

        if lora_checkpoints is not None:
            self.load_lora_weights(
                lora_checkpoints,
                lora_weights=lora_weights,
                lora_alphas=lora_alphas,
                save_base_state=False,
            )

        self._enable_cpu_offload = enable_cpu_offload
        self._enable_xformers = enable_xformers

        if self._enable_cpu_offload and self._device != "cpu":
            self.pipeline.enable_model_cpu_offload(self._device)
        else:
            self.to(device=self._device)

        if self._enable_xformers and self._device != "cpu":
            assert is_xformers_available(), "Please install xformers first."
            self.pipeline.enable_xformers_memory_efficient_attention()

    @classmethod
    @add_default_section_for_init("core/fastapi/pipeline/wan/image2video")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = None,
        config_path: Optional[str] = None,
        text_config_path: Optional[str] = None,
        image_config_path: Optional[str] = None,
        vae_config_path: Optional[str] = None,
        scheduler_config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        image_process_config_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = None,
        pretrained_lora_names: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights_path: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights: Optional[Union[float, List[float]]] = None,
        pretrained_lora_alphas: Optional[Union[float, List[float]]] = None,
        **kwargs,
    ):
        config.set_default_section("core/fastapi/pipeline/wan/image2video")
        pretrained_name = pretrained_name or config.getoption(
            "pretrained_name", "wan-v2.1-i2v-14b-480p"
        )
        pretrained_infos = nested_dict_value(pretrained_stable_infos, pretrained_name)

        config_path = config_path or config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_infos, "transformer", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = text_config_path or config.getoption(
            "text_config_path", None
        )
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrained_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        image_config_path = image_config_path or config.getoption(
            "image_config_path", None
        )
        image_config_path = pop_value(
            image_config_path,
            nested_dict_value(pretrained_infos, "image", "config"),
        )
        image_config_path = cached_path(image_config_path)

        vae_config_path = vae_config_path or config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrained_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = scheduler_config_path or config.getoption(
            "scheduler_config_path", None
        )
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrained_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        vocab_path = vocab_path or config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_infos, "text", "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        image_process_config_path = image_process_config_path or config.getoption(
            "image_process_config_path", None
        )
        image_process_config_path = pop_value(
            image_process_config_path,
            nested_dict_value(pretrained_infos, "image", "vision_config"),
        )
        image_process_config_path = cached_path(image_process_config_path)

        quant_config_path = quant_config_path or config.getoption(
            "quant_config_path", None
        )
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        weight_path = pretrained_weight_path or config.getoption(
            "pretrained_weight_path", None
        )
        device = config.getoption("device", "cpu") if device is None else device
        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        enable_xformers = config.getoption("enable_xformers", False)

        state_dict = None
        if weight_path is None and pretrained_infos is not None:
            state_dict = [
                load_weight(
                    nested_dict_value(pretrained_infos, "transformer", "weight"),
                    prefix_keys={"": "transformer."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "text", "weight"),
                    prefix_keys={"": "text."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "image", "weight"),
                    prefix_keys={"": "image."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "vae", "weight"),
                    prefix_keys={"": "vae."},
                ),
            ]

        pretrained_lora_names = pretrained_lora_names or config.getoption(
            "pretrained_lora_names", None
        )
        pretrained_lora_weights = pretrained_lora_weights or config.getoption(
            "pretrained_lora_weights", 1.0
        )
        pretrained_lora_alphas = pretrained_lora_alphas or config.getoption(
            "pretrained_lora_alphas", 32.0
        )

        if (
            isinstance(pretrained_lora_names, str)
            and pretrained_lora_weights_path is None
        ):
            pretrained_lora_weights_path = nested_dict_value(
                pretrained_stable_extensions_infos,
                pretrained_lora_names,
                "lora",
                "weight",
            )
        elif (
            isinstance(pretrained_lora_names, list)
            and pretrained_lora_weights_path is None
        ):
            pretrained_lora_weights_path = [
                nested_dict_value(
                    pretrained_stable_extensions_infos, name, "lora", "weight"
                )
                for name in pretrained_lora_names
            ]
            assert len(pretrained_lora_weights_path) == len(pretrained_lora_weights)
            assert len(pretrained_lora_weights_path) == len(pretrained_lora_alphas)

        lora_weights_path = pretrained_lora_weights_path or config.getoption(
            "pretrained_lora_weights_path", None
        )

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            image_config_path=image_config_path,
            image_process_config_path=image_process_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            vocab_path=vocab_path,
            quant_config_path=quant_config_path,
            weight_path=weight_path,
            state_dict=state_dict,
            lora_checkpoints=lora_weights_path,
            lora_weights=pretrained_lora_weights,
            lora_alphas=pretrained_lora_alphas,
            device=device,
            enable_cpu_offload=enable_cpu_offload,
            enable_xformers=enable_xformers,
        )
        return inst

    @torch.no_grad()
    @autocast(
        device_type=("cuda" if torch.cuda.is_available() else "cpu"),
        dtype=(torch.bfloat16 if is_bfloat16_available() else torch.float32),
    )
    @add_default_section_for_function("core/fastapi/pipeline/wan/image2video")
    def __call__(
        self,
        text: str,
        image: Image.Image,
        neg_text: Optional[str] = "",
        num_frames: Optional[int] = 81,
        num_fps: Optional[int] = 16,
        guidance_scale: Optional[float] = 5.0,
        strength: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
    ):
        image = image.convert("RGB")
        inputs = self.processor.image2video_inputs(
            text,
            image,
            negative_prompt=neg_text,
            max_seq_length=77,
        )
        self.seed = seed

        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self.device) if v is not None else v
            for k, v in inputs.items()
        }

        prompt_outputs = self.get_prompt_outputs(
            input_ids=inputs["input_ids"],
            negative_input_ids=inputs["negative_input_ids"],
            attention_mask=inputs["attention_mask"],
            negative_attention_mask=inputs["negative_attention_mask"],
            enable_cpu_offload=self._enable_cpu_offload,
            cpu_offload_device=self._device,
        )

        if self._enable_cpu_offload:
            self.image.to(device=self._device)

        condition_hidden_states = self.image(
            inputs["condition_pixel_values"].to(device=self._device),
            output_hidden_states=True,
        ).hidden_states[-2]

        if self._enable_cpu_offload:
            self.image.to(device="cpu")
            condition_hidden_states = condition_hidden_states.to("cpu")

        outputs = self.pipeline(
            image=inputs["vae_pixel_values"],
            prompt_embeds=prompt_outputs.prompt_embeds,
            negative_prompt_embeds=prompt_outputs.negative_prompt_embeds,
            image_embeds=condition_hidden_states,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(seed),
            num_inference_steps=num_timesteps,
            height=inputs["vae_pixel_values"].size(-2),
            width=inputs["vae_pixel_values"].size(-1),
            num_frames=num_frames,
            guidance_scale=guidance_scale,
            output_type="pt",
        )

        frames = tensor2vid(outputs.frames.float())
        name = export_to_video(frames, fps=num_fps)
        return name


@register_fastapi("core/fastapi/wan/image2video")
class WanForImage2VideoFastAPI(GenericFastAPI):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section(f"core/fastapi/wan/image2video")
        router = config.getoption("router", "/core/fastapi/wan/image2video")
        self._pipe = None
        self._router = APIRouter(prefix=router)
        self._router.add_api_route("/generate", self.serve, methods=["POST"])
        self._router.add_api_route("/status", self.status, methods=["GET"])
        self._router.add_api_route("/start", self.start, methods=["POST"])
        self._router.add_api_route("/stop", self.stop, methods=["GET"])
        self._lock = asyncio.Lock()

    @property
    def router(self):
        return self._router

    def start(
        self,
        pretrained_name: Optional[str] = "wan-v2.1-i2v-14b-480p",
        pretrained_lora_names: Optional[Union[str, List[str]]] = None,
        pretrained_lora_weights: Optional[Union[float, List[float]]] = 1.0,
        pretrained_lora_alphas: Optional[Union[float, List[float]]] = 32.0,
    ):
        self._pipe = WanForImage2VideoFastAPIPipeline.from_core_configure(
            self.config,
            pretrained_name=pretrained_name,
            pretrained_lora_names=pretrained_lora_names,
            pretrained_lora_weights=pretrained_lora_weights,
            pretrained_lora_alphas=pretrained_lora_alphas,
        )
        return "start success"

    def stop(self):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        return "stop success"

    def status(self):
        return "running" if self._pipe is not None else "stopped"

    async def serve(
        self,
        text: str,
        image: UploadFile,
        neg_text: Optional[str] = "",
        num_frames: Optional[int] = 81,
        num_fps: Optional[int] = 16,
        guidance_scale: Optional[float] = 5.0,
        strength: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
    ):
        assert self._pipe is not None
        image_bytes = await image.read()
        image = Image.open(io.BytesIO(image_bytes))
        async with self._lock:
            video = self._pipe(
                text,
                image,
                neg_text=neg_text,
                num_frames=num_frames,
                num_fps=num_fps,
                guidance_scale=guidance_scale,
                strength=strength,
                num_timesteps=num_timesteps,
                seed=seed,
            )
        buffer = io.BytesIO()
        with open(video, "rb") as f:
            buffer.write(f.read())
        buffer.seek(0)
        return StreamingResponse(
            buffer,
            media_type="video/mp4",
            headers={"Content-Disposition": "attachment; filename=output.mp4"},
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import time
import logging
import http.server
import zipfile
import tempfile
import cgi
from threading import Thread
from urllib.parse import parse_qs, urlparse
from unitorch.cli import CoreConfigureParser
from unitorch.cli import register_service, GenericService


def get_zipfile(zfs, res, idx, step):
    """
    Extracts a subset of zip files from the given list.

    Args:
        zfs (list): List of zip files.
        res (list): Result list to store extracted zip files.
        idx (int): Starting index.
        step (int): Step size.

    Returns:
        None
    """
    for i in range(idx, len(zfs), step):
        try:
            zf = zipfile.ZipFile(zfs[i])
            res[i] = zf.namelist()
        except Exception as e:
            logging.error(f"Failed to extract {zfs[i]}: {e}")
            os.remove(zfs[i])
            res[i] = None
        finally:
            pass


def get_zipfiles(zipfiles, num_thread=48):
    """
    Extracts multiple zip files using multiple threads.

    Args:
        zipfiles (list): List of zip file paths.
        num_thread (int): Number of threads to use.

    Returns:
        list: Extracted zip files.
    """
    num_thread = min(len(zipfiles), num_thread)
    threads = [None] * num_thread
    results = [None] * len(zipfiles)
    for i in range(len(threads)):
        threads[i] = Thread(
            target=get_zipfile,
            args=(zipfiles, results, i, num_thread),
        )
        threads[i].start()

    for i in range(len(threads)):
        threads[i].join()
    return results


class ZipSaverServer(http.server.BaseHTTPRequestHandler):
    zip_set = {}
    zip_file_prefix = "zip_files"
    zip_folder = None
    curr_file = None
    next_zip_file_index = 0
    num_files_in_zip = 0
    max_files_per_zip = 10000

    @classmethod
    def create_new_file(cls):
        if cls.curr_file is not None:
            cls.curr_file.close()
        cls.num_files_in_zip = 0
        name = os.path.join(
            cls.zip_folder, f"{cls.zip_file_prefix}_{cls.next_zip_file_index}.zip"
        )
        cls.curr_file = zipfile.ZipFile(name, "w")
        cls.next_zip_file_index += 1

    def do_POST(self):
        content_type, pdict = cgi.parse_header(self.headers["Content-Type"])
        name = parse_qs(urlparse(self.path).query).get("name", [None])[0]
        if content_type == "multipart/form-data":
            pdict["boundary"] = bytes(pdict["boundary"], "utf-8")
            fields = cgi.parse_multipart(self.rfile, pdict)
            if "file" in fields:
                content = fields["file"][0]
                if name in ZipSaverServer.zip_set:
                    self.send_response(200)
                    self.end_headers()
                else:
                    if (
                        ZipSaverServer.num_files_in_zip
                        >= ZipSaverServer.max_files_per_zip
                    ):
                        ZipSaverServer.create_new_file()
                    ZipSaverServer.curr_file.writestr(name, content)
                    ZipSaverServer.num_files_in_zip += 1
                    ZipSaverServer.zip_set.add(name)
                    self.send_response(200)
                    self.send_header("Content-type", "application/zip")
                    self.end_headers()
            else:
                self.send_response(400)
                self.end_headers()
        else:
            self.send_response(400)
            self.end_headers()


@register_service("core/service/zip_saver")
class ZipSaverService(GenericService):
    def __init__(self, config: CoreConfigureParser):
        self.config = config
        config.set_default_section("core/service/zip_saver")
        self.ip = config.getoption("ip", "0.0.0.0")
        self.port = config.getoption("port", 11231)
        self.name = config.getoption("processname", "core_zip_saver_service")
        self.zip_folder = config.getoption("zip_folder", None)
        self.zip_file_prefix = config.getoption("zip_file_prefix", "zip_files")
        self.zip_extension = config.getoption("zip_extension", ".zip")
        self.max_files_per_zip = config.getoption("max_files_per_zip", 10000000)
        assert self.zip_folder is not None
        if not os.path.exists(self.zip_folder):
            os.makedirs(self.zip_folder)

        self.num_thread = config.getoption("num_thread", 20)
        if isinstance(self.zip_folder, str):
            self.zip_folder = os.path.abspath(self.zip_folder)
            zip_files = [
                os.path.join(self.zip_folder, f)
                for f in os.listdir(self.zip_folder)
                if f.endswith(self.zip_extension)
            ]
        elif isinstance(self.zip_folder, list):
            self.zip_folder = [os.path.abspath(f) for f in self.zip_folder]
            zip_files = []
            for folder in list(set(self.zip_folder)):
                zip_files += [
                    os.path.join(folder, f)
                    for f in os.listdir(folder)
                    if f.endswith(self.zip_extension)
                ]

        zip_data = get_zipfiles(zip_files, self.num_thread)
        self.zip_set = set()
        for data in zip_data:
            if data is None:
                continue
            self.zip_set.update(data)
        self.next_zip_file_index = (
            max(
                [
                    int(f.split("_")[-1].split(".")[0])
                    for f in os.listdir(self.zip_folder)
                    if f.startswith(self.zip_file_prefix)
                ],
                default=-1,
            )
            + 1
        )

    def start(self, **kwargs):
        ZipSaverServer.zip_folder = self.zip_folder
        ZipSaverServer.zip_set = self.zip_set
        ZipSaverServer.max_files_per_zip = self.max_files_per_zip
        ZipSaverServer.zip_file_prefix = self.zip_file_prefix
        ZipSaverServer.next_zip_file_index = self.next_zip_file_index + 1
        name = os.path.join(
            self.zip_folder, f"{self.zip_file_prefix}_{self.next_zip_file_index}.zip"
        )
        ZipSaverServer.curr_file = name
        ZipSaverServer.curr_file = zipfile.ZipFile(name, "w")
        self.httpd = http.server.HTTPServer((self.ip, self.port), ZipSaverServer)
        self.httpd.serve_forever()

    def stop(self, **kwargs):
        ZipSaverServer.curr_file.close()

    def restart(self, **kwargs):
        pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli import hf_endpoint_url

# pretrained infos
pretrained_llava_infos = {
    "llava-v1.6-mistral-7b-hf": {
        "config": hf_endpoint_url(
            "/llava-hf/llava-v1.6-mistral-7b-hf/resolve/main/config.json"
        ),
        "vocab": hf_endpoint_url(
            "/llava-hf/llava-v1.6-mistral-7b-hf/resolve/main/tokenizer.model"
        ),
        "vision_config": hf_endpoint_url(
            "/llava-hf/llava-v1.6-mistral-7b-hf/resolve/main/preprocessor_config.json"
        ),
        "weight": [
            hf_endpoint_url(
                f"/llava-hf/llava-v1.6-mistral-7b-hf/resolve/main/model-{str(i).rjust(5, '0')}-of-00004.safetensors"
            )
            for i in range(1, 5)
        ],
    },
    "llava-v1.6-joycaption-2": {
        "config": hf_endpoint_url(
            "/fancyfeast/llama-joycaption-alpha-two-hf-llava/resolve/main/config.json"
        ),
        "tokenizer": hf_endpoint_url(
            "/fancyfeast/llama-joycaption-alpha-two-hf-llava/resolve/main/tokenizer.json"
        ),
        "vision_config": hf_endpoint_url(
            "/fancyfeast/llama-joycaption-alpha-two-hf-llava/resolve/main/preprocessor_config.json"
        ),
        "weight": [
            hf_endpoint_url(
                f"/fancyfeast/llama-joycaption-alpha-two-hf-llava/resolve/main/model-{str(i).rjust(5, '0')}-of-00004.safetensors"
            )
            for i in range(1, 5)
        ],
    },
}

pretrained_llava_extensions_infos = {}

import unitorch.cli.models.llava.modeling
import unitorch.cli.models.llava.processing
from unitorch.cli.models.llava.modeling import (
    LlavaMistralClipForClassification,
    LlavaMistralClipForGeneration,
)
from unitorch.cli.models.llava.processing import LlavaMistralClipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from hmac import new
import os
import torch
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torchvision.transforms import Resize, CenterCrop, ToTensor, Normalize, Compose
from transformers import BertTokenizer, GroundingDinoImageProcessor
from unitorch.utils import resize_shortest_edge, pop_value
from unitorch.models import (
    HfTextClassificationProcessor,
    HfImageClassificationProcessor,
    GenericOutputs,
)


class GroundingDinoProcessor(
    HfTextClassificationProcessor, HfImageClassificationProcessor
):
    def __init__(
        self,
        vocab_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 128,
        position_start_id: Optional[int] = 0,
    ):
        """
        Args:
            vision_config_path: vision config path to detr processor
            min_size_test: resize shortest edge parameters
            max_size_test: resize shortest edge parameters
        """
        self.bert_tokenizer = BertTokenizer(
            vocab_path, do_basic_tokenize=True, do_lower_case=True
        )
        self.vision_processor = GroundingDinoImageProcessor.from_json_file(
            vision_config_path
        )

        HfTextClassificationProcessor.__init__(
            self,
            tokenizer=self.bert_tokenizer,
            max_seq_length=max_seq_length,
            source_type_id=0,
            target_type_id=1,
            position_start_id=position_start_id,
        )

        HfImageClassificationProcessor.__init__(
            self,
            vision_processor=self.vision_processor,
        )

    def detection(
        self,
        text: str,
        image: Union[str, Image.Image],
        bboxes: List[List[float]],
        classes: List[str],
    ):
        """
        Args:
            image: input image
            bboxes: bboxes to image
            classes: class to each bbox
        """
        if isinstance(image, str):
            image = Image.open(image).convert("RGB")
        org_w, org_h = image.size
        pixel_outputs = HfImageClassificationProcessor.classification(self, image)
        text_outputs = HfTextClassificationProcessor.classification(self, text)

        bboxes = torch.tensor(bboxes).float()
        bboxes[:, 0] = bboxes[:, 0] / org_w
        bboxes[:, 1] = bboxes[:, 1] / org_h
        bboxes[:, 2] = bboxes[:, 2] / org_w
        bboxes[:, 3] = bboxes[:, 3] / org_h

        assert all(c in text for c in classes)
        ground_truth = text_outputs.input_ids.long().tolist()
        class_ids = []
        for c in classes:
            class_tokens = self.tokenizer.tokenize(c)
            class_token_ids = self.tokenizer.convert_tokens_to_ids(class_tokens)
            class_idx = ground_truth.index(class_token_ids[0])
            class_ids.append(class_idx)

        classes = torch.tensor(class_ids)
        if bboxes.dim() == 1:
            bboxes = bboxes.unsqueeze(0)

        assert (
            bboxes.size(-1) == 4 and classes.dim() == 1 and len(classes) == len(bboxes)
        )
        return GenericOutputs(
            pixel_values=pixel_outputs.pixel_values,
            input_ids=text_outputs.input_ids,
            attention_mask=text_outputs.attention_mask,
            token_type_ids=text_outputs.token_type_ids,
            bboxes=bboxes,
            classes=classes,
        )

    def detection_inputs(
        self,
        text: str,
        image: Union[str, Image.Image],
    ):
        """
        Args:
            image: input image
        """
        if isinstance(image, str):
            image = Image.open(image).convert("RGB")
        pixel_outputs = HfImageClassificationProcessor.classification(self, image)
        text_outputs = HfTextClassificationProcessor.classification(self, text)
        return GenericOutputs(
            pixel_values=pixel_outputs.pixel_values,
            input_ids=text_outputs.input_ids,
            attention_mask=text_outputs.attention_mask,
            token_type_ids=text_outputs.token_type_ids,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image, ImageOps
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import nested_dict_value
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
)
from unitorch.cli.pipelines.stable import StableForImageInpaintingPipeline
from unitorch.cli.pipelines.tools import controlnet_processes
from unitorch.cli.webuis import (
    supported_scheduler_names,
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_controlnet_layout,
    create_lora_layout,
)
from unitorch.cli.webuis import SimpleWebUI


class StableImageInpaintingWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_stable_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, ["^stable-v1.5-", "^stable-v2-", "^stable-v2.1"]
    )
    pretrained_extension_names = list(pretrained_stable_extensions_infos.keys())
    supported_controlnet_names = matched_pretrained_names(
        pretrained_extension_names,
        [
            "^stable-v1.5-controlnet-",
            "^stable-v2-controlnet-",
            "^stable-v2.1-controlnet-",
        ],
    )
    supported_controlnet_process_names = list(controlnet_processes.keys())
    supported_lora_names = matched_pretrained_names(
        pretrained_extension_names,
        ["^stable-v1.5-lora-", "^stable-v2-lora-", "^stable-v2.1-lora-"],
    )
    supported_schedulers = supported_scheduler_names

    def __init__(self, config: CoreConfigureParser):
        self._pipe = None
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        prompt = create_element(
            "text", "Input Prompt", lines=3, placeholder="Prompt", show_label=False
        )
        negative_prompt = create_element(
            "text",
            "Input Negative Prompt",
            lines=3,
            placeholder="Negative Prompt",
            show_label=False,
        )
        scheduler = create_element(
            "dropdown",
            "Sampling Method",
            values=self.supported_schedulers,
            default=self.supported_schedulers[0],
        )
        steps = create_element(
            "slider", "Diffusion Steps", min_value=1, max_value=100, step=1, default=25
        )
        height = create_element(
            "slider", "Image Height", min_value=1, max_value=2048, step=1, default=1024
        )
        width = create_element(
            "slider", "Image Width", min_value=1, max_value=2048, step=1, default=1024
        )
        image = create_element("image_editor", "Input Image")
        mask_image = create_element("image", "Input Image Mask")

        guidance_scale = create_element(
            "slider", "Guidance Scale", min_value=0, max_value=50, step=0.1, default=7.5
        )
        strength = create_element(
            "slider", "Strength", min_value=0, max_value=1, step=0.01, default=0.8
        )

        seed = create_element(
            "slider", "Seed", min_value=0, max_value=9999, step=1, default=42
        )

        ## extensions
        self.num_controlnets = 5
        controlnet_layout_group = create_controlnet_layout(
            self.supported_controlnet_names,
            self.supported_controlnet_process_names,
            num_controlnets=self.num_controlnets,
        )
        controlnets = controlnet_layout_group.controlnets
        controlnet_layout = controlnet_layout_group.layout

        controlnet_params = []
        for controlnet in controlnets:
            controlnet_params += [
                controlnet.checkpoint,
                controlnet.output_image,
                controlnet.guidance_scale,
            ]

        self.num_loras = 5
        lora_layout_group = create_lora_layout(
            self.supported_lora_names, num_loras=self.num_loras
        )
        loras = lora_layout_group.loras
        lora_layout = lora_layout_group.layout
        lora_params = []
        for lora in loras:
            lora_params += [
                lora.checkpoint,
                lora.weight,
                lora.alpha,
                lora.url,
                lora.file,
            ]

        generate = create_element("button", "Generate", variant="primary", scale=2)
        output_image = create_element("image", "Output Image")

        # create layouts
        top1 = create_column(pretrain_layout)
        top2 = create_row(
            create_column(prompt, negative_prompt, scale=4),
            create_column(generate),
        )
        left_generation = create_tab(
            create_row(image, mask_image),
            create_row(scheduler, steps),
            create_row(height, width),
            create_row(guidance_scale, strength),
            create_row(seed),
            name="Generation",
        )
        left_extension = create_tab(
            create_row(controlnet_layout),
            create_row(lora_layout),
            name="Extensions",
        )
        left_settings = create_tab(
            name="Settings",
        )

        left = create_tabs(left_generation, left_extension, left_settings)
        right = create_column(output_image)
        iface = create_blocks(top1, top2, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(fn=self.start, inputs=[name], outputs=[status], trigger_mode="once")
        stop.click(fn=self.stop, outputs=[status], trigger_mode="once")

        image.change(fn=self.composite_images, inputs=[image], outputs=[mask_image])

        for controlnet in controlnets:
            controlnet.input_image.upload(
                fn=self.processing_controlnet_inputs,
                inputs=[controlnet.input_image, controlnet.process],
                outputs=[controlnet.output_image],
            )
            controlnet.process.change(
                fn=self.processing_controlnet_inputs,
                inputs=[controlnet.input_image, controlnet.process],
                outputs=[controlnet.output_image],
            )

        for lora in loras:
            lora.checkpoint.change(
                fn=lambda x: nested_dict_value(
                    pretrained_stable_extensions_infos, x, "text"
                ),
                inputs=[lora.checkpoint],
                outputs=[lora.text],
            )

        generate.click(
            fn=self.serve,
            inputs=[
                prompt,
                image,
                mask_image,
                negative_prompt,
                height,
                width,
                guidance_scale,
                strength,
                steps,
                seed,
                scheduler,
                *controlnet_params,
                *lora_params,
            ],
            outputs=[output_image],
            trigger_mode="once",
        )
        image.change(
            lambda x: x["background"].size,
            inputs=[image],
            outputs=[width, height],
        )
        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="Inpainting", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._name == pretrained_name and self._status == "Running":
            return self._status
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = StableForImageInpaintingPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def composite_images(self, images):
        layers = images["layers"]
        if len(layers) == 0:
            return None
        image = layers[0]
        for i in range(1, len(layers)):
            image = Image.alpha_composite(image, layers[i])
        image = image.convert("L")
        image = image.point(lambda p: p < 5 and 255)
        image = ImageOps.invert(image)
        return image

    def processing_controlnet_inputs(self, image, process):
        pfunc = controlnet_processes.get(process, None)
        if pfunc is not None and image is not None:
            return pfunc(image)
        return image

    def serve(
        self,
        text: str,
        image: Image.Image,
        mask_image: Image.Image,
        negative_text: str,
        height: int,
        width: int,
        guidance_scale: float,
        strength: float,
        num_timesteps: int,
        seed: int,
        scheduler: str,
        *params,
    ):
        assert self._pipe is not None
        controlnet_params = params[: self.num_controlnets * 3]
        lora_params = params[self.num_controlnets * 3 :]
        controlnet_checkpoints = controlnet_params[::3]
        controlnet_images = controlnet_params[1::3]
        controlnet_guidance_scales = controlnet_params[2::3]
        lora_checkpoints = lora_params[::5]
        lora_weights = lora_params[1::5]
        lora_alphas = lora_params[2::5]
        lora_urls = lora_params[3::5]
        lora_files = lora_params[4::5]
        image = self._pipe(
            text,
            image["background"],
            mask_image,
            negative_text,
            width=width,
            height=height,
            guidance_scale=guidance_scale,
            strength=strength,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            controlnet_checkpoints=controlnet_checkpoints,
            controlnet_images=controlnet_images,
            controlnet_guidance_scales=controlnet_guidance_scales,
            lora_checkpoints=lora_checkpoints,
            lora_weights=lora_weights,
            lora_alphas=lora_alphas,
            lora_urls=lora_urls,
            lora_files=lora_files,
        )
        return image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import re
import io
import torch
import torch.nn as nn
import requests
import hashlib
import imageio
import tempfile
import logging
import numpy as np
import safetensors
from PIL import Image
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil, pt_to_pil, export_to_gif
from unitorch import is_opencv_available
from unitorch.utils import load_weight, tensor2vid
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli.models.modeling_utils import TensorsOutputs, TensorsTargets

from unitorch.cli import cached_path


def numpy2vid(video: np.ndarray) -> List[np.ndarray]:
    if video.ndim == 4:
        video = np.expand_dims(video, axis=0)
    i, f, c, h, w = video.shape
    images = np.transpose(video, (1, 3, 0, 4, 2))
    images = np.transpose(video, (1, 3, 0, 4, 2))
    # Reshape to (frames, height, batch_size * width, channels)
    images = images.reshape(f, h, i * w, c)
    # Convert to uint8 and scale pixel values
    images = [(frame * 255).astype("uint8") for frame in images]
    return images


def export_to_video(
    video_frames: List[np.ndarray],
    output_video_path: str = None,
    fps: int = 8,
) -> str:
    if output_video_path is None:
        output_video_path = tempfile.NamedTemporaryFile(suffix=".mp4").name

    video_writer = imageio.get_writer(output_video_path, fps=fps)
    for frame in video_frames:
        video_writer.append_data(frame)
    video_writer.close()
    return output_video_path


@dataclass
class DiffusionOutputs(TensorsOutputs, WriterMixin):
    outputs: torch.Tensor


@dataclass
class DiffusionTargets(TensorsTargets):
    targets: torch.Tensor
    masks: Optional[torch.Tensor] = torch.empty(0)


class DiffusionProcessor:
    def __init__(
        self,
        output_folder: Optional[str] = None,
        http_url: Optional[str] = None,
        video_fps: Optional[int] = 8,
    ):
        assert output_folder is not None or http_url is not None
        self.output_folder = output_folder
        if output_folder is not None and not os.path.exists(output_folder):
            os.makedirs(self.output_folder, exist_ok=True)
        self.http_url = http_url
        self.video_fps = video_fps
        if self.output_folder is None:
            self.output_folder = tempfile.mkdtemp()

    @classmethod
    @add_default_section_for_init("core/process/diffusion")
    def from_core_configure(cls, config, **kwargs):
        pass

    def save_image(self, image: Image.Image):
        md5 = hashlib.md5()
        md5.update(image.tobytes())
        name = md5.hexdigest() + ".jpg"
        if self.http_url is not None:
            byte = io.BytesIO()
            image.save(byte, format="JPEG")
            for _ in range(3):
                resp = requests.post(
                    self.http_url.format(name), files={"file": byte.getvalue()}
                )
                if resp.status_code == 200:
                    break
            if resp.status_code != 200:
                logging.error(f"Failed to save image {name} to zip.")
        else:
            output_path = f"{self.output_folder}/{name}"
            image.save(output_path)
        return name

    def save_gif(self, images: List[Image.Image]):
        md5 = hashlib.md5()
        for image in images:
            md5.update(image.tobytes())
        name = md5.hexdigest() + ".gif"
        output_path = f"{self.output_folder}/{name}"
        export_to_gif(images, output_path)
        if self.http_url is not None:
            byte = open(output_path, "rb")
            for _ in range(3):
                resp = requests.post(self.http_url.format(name), files={"file": byte})
                if resp.status_code == 200:
                    break
            if resp.status_code != 200:
                logging.error(f"Failed to save gif {name} to zip.")
        return name

    def save_video(self, video):
        md5 = hashlib.md5()
        for image in video:
            md5.update(image.tobytes())
        name = md5.hexdigest() + ".mp4"
        output_path = f"{self.output_folder}/{name}"
        export_to_video(video, output_path, fps=self.video_fps)
        if self.http_url is not None:
            byte = open(output_path, "rb")
            for _ in range(3):
                resp = requests.post(self.http_url.format(name), files={"file": byte})
                if resp.status_code == 200:
                    break
            if resp.status_code != 200:
                logging.error(f"Failed to save video {name} to zip.")
        return name

    @register_process("core/postprocess/diffusion/image")
    def _diffusion_image(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        images = outputs.outputs.numpy()
        images = numpy_to_pil(images)
        results["diffusion"] = [self.save_image(image) for image in images]
        return WriterOutputs(results)

    @register_process("core/postprocess/diffusion/gif")
    def _diffusion_gif(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        images = outputs.outputs.numpy()
        images = [numpy_to_pil(image) for image in images]
        results["diffusion"] = [self.save_gif(image) for image in images]
        return WriterOutputs(results)

    @register_process("core/postprocess/diffusion/video")
    def _diffusion_video(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        videos = outputs.outputs.numpy()
        videos = [numpy2vid(video) for video in videos]
        results["diffusion"] = [self.save_video(video) for video in videos]
        return WriterOutputs(results)


def diffusion_model_decorator(cls):
    class DiffusionModel(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            if "__diffusion_model__" in kwargs:
                self.model = kwargs.pop("__diffusion_model__")
            else:
                self.model = cls(*args, **kwargs)

            __more_attrs__ = [
                "load_state_dict",
                "state_dict",
                "save_checkpoint",
                "from_checkpoint",
                "from_pretrained",
            ]
            for __more_attr__ in __more_attrs__:
                setattr(self, __more_attr__, getattr(self.model, __more_attr__))

            assert hasattr(self.model, "generate")

            self.model.register_forward_hook(self._hook)
            self.__in_training__ = False

        def _hook(self, module, inputs, outputs):
            self.__in_training__ = True

        def forward(self, *args, **kwargs):
            if self.training or self.__in_training__:
                return self.model(*args, **kwargs)
            return self.model.generate(*args, **kwargs)

        @classmethod
        def from_core_configure(_cls, cfg, **kwargs):
            model = cls.from_core_configure(cfg, **kwargs)
            return _cls(__diffusion_model__=model)

    return DiffusionModel



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import math
import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, Iterator
from torch.utils.data import (
    BatchSampler,
    DataLoader,
    Dataset,
    RandomSampler,
    SequentialSampler,
)
from torch.utils.data.distributed import DistributedSampler


def get_local_rank() -> int:
    """
    Get the local rank from the environment variable 'LOCAL_RANK'.

    Returns:
        int: Local rank or -1 if not found.
    """
    return int(os.environ.get("LOCAL_RANK", -1))


class DistributedSkipSampler(DistributedSampler):
    def __init__(
        self,
        dataset: Dataset,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        shuffle: Optional[bool] = True,
        seed: Optional[int] = 0,
        drop_last: Optional[bool] = False,
        skip_step: Optional[int] = 0,
    ) -> None:
        super().__init__(
            dataset=dataset,
            num_replicas=num_replicas,
            rank=rank,
            shuffle=shuffle,
            seed=seed,
            drop_last=drop_last,
        )
        self.skip_step = skip_step

    def __iter__(self):
        if self.shuffle:
            # Deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        if not self.drop_last:
            # Add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # Remove tail of data to make it evenly divisible
            indices = indices[: self.total_size]
        assert len(indices) == self.total_size

        # Subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices[self.skip_step :])

    def __len__(self) -> int:
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        self.epoch = epoch

    def set_skip_step(self, step: int) -> None:
        self.skip_step = step


class RandomSkipSampler(RandomSampler):
    def __init__(
        self,
        data_source: Dataset,
        replacement: Optional[bool] = False,
        num_samples: Optional[int] = None,
        skip_step: Optional[int] = 0,
    ):
        super().__init__(
            data_source=data_source,
            replacement=replacement,
            num_samples=num_samples,
        )
        self.skip_step = skip_step

    def __iter__(self):
        n = len(self.data_source)
        if self.replacement:
            return iter(
                torch.randint(
                    high=n,
                    size=(self.num_samples,),
                    dtype=torch.int64,
                ).tolist()[self.skip_step :]
            )
        return iter(torch.randperm(n).tolist()[self.skip_step :])

    def set_skip_step(self, step: int) -> None:
        self.skip_step = step


class SequentialSkipSampler(SequentialSampler):
    def __init__(
        self,
        data_source: Dataset,
        skip_step: Optional[int] = 0,
    ):
        super().__init__(
            data_source=data_source,
        )
        self.skip_step = skip_step

    def __iter__(self):
        return iter(range(len(self.data_source))[self.skip_step :])

    def set_skip_step(self, step: int) -> None:
        self.skip_step = step



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from transformers import T5Tokenizer
from unitorch.models import HfTextGenerationProcessor


def get_t5_tokenizer(
    vocab_path,
    special_input_ids=dict(),
):
    tokenizer = T5Tokenizer(vocab_path)
    for token, _id in special_input_ids.items():
        tokenizer.added_tokens_encoder[token] = _id
        tokenizer.unique_no_split_tokens.append(token)
        tokenizer.added_tokens_decoder[_id] = token
        tokenizer.add_tokens(token, special_tokens=True)
    return tokenizer


class T5Processor(HfTextGenerationProcessor):
    """
    Processor for T5-based text generation models.
    """

    def __init__(
        self,
        vocab_path: str,
        special_input_ids: Optional[Dict] = dict(),
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
    ):
        """
        Initializes the T5Processor.

        Args:
            vocab_path (str): Path to the vocabulary file.
            special_input_ids (Optional[Dict]): Special input IDs. Defaults to an empty dictionary.
            max_seq_length (Optional[int]): Maximum sequence length. Defaults to 128.
            max_gen_seq_length (Optional[int]): Maximum generated sequence length. Defaults to 48.
        """
        tokenizer = get_t5_tokenizer(
            vocab_path,
            special_input_ids=special_input_ids,
        )
        tokenizer.bos_token_id = 0
        tokenizer.bos_token = tokenizer.convert_ids_to_tokens(0)
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.sep_token_id = tokenizer.eos_token_id
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import random
import logging
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch import autocast
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers.models.mask2former.modeling_mask2former import (
    Mask2FormerConfig,
    Mask2FormerModel,
    Mask2FormerLoss,
)
from unitorch.models import GenericModel, GenericOutputs


class Mask2FormerForSegmentation(GenericModel):
    def __init__(
        self,
        config_path: str,
    ):
        """
        Initializes a SamForSegmentation model for segmentation tasks.

        Args:
            config_path (str): The path to the Sam Transformer configuration file.
        """
        super().__init__()
        config = Mask2FormerConfig.from_json_file(config_path)

        self.model = Mask2FormerModel(config)
        self.weight_dict: Dict[str, float] = {
            "loss_cross_entropy": config.class_weight,
            "loss_mask": config.mask_weight,
            "loss_dice": config.dice_weight,
        }

        self.class_predictor = nn.Linear(config.hidden_dim, config.num_labels + 1)
        self.criterion = Mask2FormerLoss(config=config, weight_dict=self.weight_dict)
        self.init_weights()

    def forward(self):
        """
        Performs a forward pass of the SamForSegmentation model.
        """
        raise NotImplementedError

    def segment(
        self,
        pixel_values: torch.Tensor,
        pixel_mask: Optional[torch.Tensor] = None,
    ):
        """
        Performs a forward pass of the SamForSegmentation model.
        """
        outputs = self.model(
            pixel_values=pixel_values,
            pixel_mask=pixel_mask,
            output_hidden_states=True,
        )

        decoder_output = outputs.transformer_decoder_intermediate_states[-1]
        class_queries_logits = self.class_predictor(decoder_output.transpose(0, 1))

        masks_queries_logits = outputs.masks_queries_logits[-1]
        masks_queries_logits = nn.functional.interpolate(
            masks_queries_logits,
            scale_factor=4,
            mode="bilinear",
            align_corners=False,
        )
        return GenericOutputs(masks=masks_queries_logits, classes=class_queries_logits)
